from __future__ import absolute_import

import six
import json

from ..hubstorage import ValueTooLarge as _ValueTooLarge
from .utils import update_kwargs
from .exceptions import ValueTooLarge


class _Proxy(object):
    """A helper to create a class instance and proxy its methods to origin.

    The internal proxy class is useful to link class attributes from its
    origin depending on the origin base class as a part of init logic:

    - :class:`~scrapinghub.hubstorage.resourcetype.ItemsResourceType` provides
        items-based attributes to access items in an arbitrary collection with
        get/write/flush/close/stats/iter methods.

    - :class:`~scrapinghub.hubstorage.resourcetype.DownloadableResource` provides
        download-based attributes to iter through collection with or without
        msgpack support.
    """

    def __init__(self, cls, client, key):
        self.key = key
        self._client = client
        self._origin = cls(client._hsclient, key)

    def list(self, *args, **kwargs):
        """Convenient shortcut to list iter results.

        Please note that :meth:`list` method can use a lot of memory and for a
        large amount of elements it's recommended to iterate through it via
        :meth:`iter` method (all params and available filters are same for both
        methods).
        """
        return list(self.iter(*args, **kwargs))

    def _modify_iter_params(self, params):
        """A helper to modify iter*() params on-the-fly.

        The method is internal and should be redefined in subclasses.

        :param params: a dictionary with input parameters.
        :return: an updated dictionary with parameters.
        :rtype: :class:`dict`
        """
        return _format_iter_filters(params)


class _ItemsResourceProxy(_Proxy):

    def get(self, key, **params):
        """Get element from collection.

        :param key: element key.
        :return: a dictionary with element data.
        :rtype: :class:`dict`
        """
        return self._origin.get(key, **params)

    def write(self, item):
        """Write new element to collection.

        :param item: element data dict to write.
        """
        try:
            return self._origin.write(item)
        except _ValueTooLarge as exc:
            raise ValueTooLarge(str(exc))

    def iter(self, _key=None, count=None, **params):
        """Iterate over elements in collection.

        :param count: limit amount of elements.
        :return: a generator object over a list of element dictionaries.
        :rtype: :class:`types.GeneratorType[dict]`
        """
        update_kwargs(params or {}, count=count)
        params = self._modify_iter_params(params)
        return self._origin.list(_key, **params)

    def flush(self):
        """Flush data from writer threads."""
        self._origin.flush()

    def stats(self):
        """Get resource stats.

        :return: a dictionary with stats data.
        :rtype: :class:`dict`
        """
        return self._origin.stats()

    def close(self, block=True):
        """Close writers one-by-one."""
        self._origin.close(block)


class _DownloadableProxyMixin(object):

    def iter(self, _path=None, count=None, requests_params=None, **apiparams):
        """A general method to iterate through elements.

        :param count: limit amount of elements.
        :return: an iterator over elements list.
        :rtype: :class:`collections.Iterable`
        """
        update_kwargs(apiparams, count=count)
        apiparams = self._modify_iter_params(apiparams)
        return self._origin.iter_values(_path, requests_params, **apiparams)


class _MappingProxy(_Proxy):
    """A helper class to support basic get/set interface for dict-like
    collections of elements.
    """

    def get(self, key):
        """Get element value by key.

        :param key: a string key
        """
        return next(self._origin.apiget(key))

    def set(self, key, value):
        """Set element value.

        :param key: a string key
        :param value: new value to set for the key
        """
        self._origin.apipost(key, data=json.dumps(value), is_idempotent=True)

    def update(self, values):
        """Update multiple elements at once.

        The method provides convenient interface for partial updates.

        :param values: a dictionary with key/values to update.
        """
        if not isinstance(values, dict):
            raise TypeError("values should be a dict")
        data = next(self._origin.apiget())
        data.update(values)
        self._origin.apipost(jl={k: v for k, v in six.iteritems(data)
                                 if k not in self._origin.ignore_fields},
                             is_idempotent=True)

    def delete(self, key):
        """Delete element by key.

        :param key: a string key
        """
        self._origin.apidelete(key)

    def iter(self):
        """Iterate through key/value pairs.

        :return: an iterator over key/value pairs.
        :rtype: :class:`collections.Iterable`
        """
        return six.iteritems(next(self._origin.apiget()))


def _format_iter_filters(params):
    """Format iter() filter param on-the-fly.

    Support passing multiple filters at once as a list with tuples.
    """
    filters = params.get('filter')
    if filters and isinstance(filters, list):
        filter_data = []
        for elem in params.pop('filter'):
            if isinstance(elem, six.string_types):
                filter_data.append(elem)
            elif isinstance(elem, (list, tuple)):
                filter_data.append(json.dumps(elem))
            else:
                raise ValueError(
                    "Filter condition must be string, tuple or list")
        if filter_data:
            params['filter'] = filter_data
    return params



from __future__ import absolute_import

from ..hubstorage.activity import Activity as _Activity
from ..hubstorage.collectionsrt import Collections as _Collections
from ..hubstorage.project import Settings as _Settings

from .activity import Activity
from .collections import Collections
from .frontiers import _HSFrontier, Frontiers
from .jobs import Jobs
from .proxy import _MappingProxy
from .spiders import Spiders
from .utils import parse_project_id


class Projects(object):
    """Collection of projects available to current user.

    Not a public constructor: use :class:`~scrapinghub.client.ScrapinghubClient`
    client instance to get a :class:`Projects` instance.
    See :attr:`scrapinghub.client.Scrapinghub.projects` attribute.

    Usage::

        >>> client.projects
        <scrapinghub.client.projects.Projects at 0x1047ada58>
    """

    def __init__(self, client):
        self._client = client

    def get(self, project_id):
        """Get project for a given project id.

        :param project_id: integer or string numeric project id.
        :return: a project object.
        :rtype: :class:`Project`

        Usage::

            >>> project = client.projects.get(123)
            >>> project
            <scrapinghub.client.projects.Project at 0x106cdd6a0>
        """
        return Project(self._client, parse_project_id(project_id))

    def list(self):
        """Get list of projects available to current user.

        :return: a list of project ids.
        :rtype: :class:`list[int]`

        Usage::

            >>> client.projects.list()
            [123, 456]
        """
        return self._client._connection.project_ids()

    def iter(self):
        """Iterate through list of projects available to current user.

        Provided for the sake of API consistency.

        :return: an iterator over project ids list.
        :rtype: :class:`collections.Iterable[int]`
        """
        return iter(self.list())

    def summary(self, state=None, **params):
        """Get short summaries for all available user projects.

        :param state: a string state or a list of states.
        :return: a list of dictionaries: each dictionary represents a project
            summary (amount of pending/running/finished jobs and a flag if it
            has a capacity to run new jobs).
        :rtype: :class:`list[dict]`

        Usage::

            >>> client.projects.summary()
            [{'finished': 674,
              'has_capacity': True,
              'pending': 0,
              'project': 123,
              'running': 1},
             {'finished': 33079,
              'has_capacity': True,
              'pending': 0,
              'project': 456,
              'running': 2}]
        """
        if state:
            params['state'] = state
        return self._client._hsclient.projects.jobsummaries(**params)


class Project(object):
    """Class representing a project object and its resources.

    Not a public constructor: use :class:`~scrapinghub.client.ScrapinghubClient`
    instance or :class:`Projects` instance to get a :class:`Project` instance.
    See :meth:`scrapinghub.client.ScrapinghubClient.get_project` or
    :meth:`Projects.get` methods.

    :ivar key: string project id.
    :ivar activity: :class:`~scrapinghub.client.activity.Activity` resource object.
    :ivar collections: :class:`~scrapinghub.client.collections.Collections` resource object.
    :ivar frontiers: :class:`~scrapinghub.client.frontiers.Frontiers` resource object.
    :ivar jobs: :class:`~scrapinghub.client.jobs.Jobs` resource object.
    :ivar settings: :class:`~scrapinghub.client.settings.Settings` resource object.
    :ivar spiders: :class:`~scrapinghub.client.spiders.Spiders` resource object.

    Usage::

        >>> project = client.get_project(123)
        >>> project
        <scrapinghub.client.projects.Project at 0x106cdd6a0>
        >>> project.key
        '123'
    """

    def __init__(self, client, project_id):
        self.key = str(project_id)
        self._client = client

        # sub-resources
        self.jobs = Jobs(client, project_id)
        self.spiders = Spiders(client, project_id)

        # proxied sub-resources
        self.activity = Activity(_Activity, client, project_id)
        self.collections = Collections(_Collections, client, project_id)
        self.frontiers = Frontiers(_HSFrontier, client, project_id)
        self.settings = Settings(_Settings, client, project_id)


class Settings(_MappingProxy):
    """Class representing job metadata.

    Not a public constructor: use :class:`Project` instance to get a
    :class:`Settings` instance. See :attr:`Project.settings` attribute.

    Usage:

    - get project settings instance::

        >>> project.settings
        <scrapinghub.client.projects.Settings at 0x10ecf1250>

    - iterate through project settings::

        >>> project.settings.iter()
        <dictionary-itemiterator at 0x10ed11578>

    - list project settings::

        >>> project.settings.list()
        [(u'default_job_units', 2), (u'job_runtime_limit', 20)]

    - get setting value by name::

        >>> project.settings.get('default_job_units')
        2

    - update setting value (some settings are read-only)::

        >>> project.settings.set('default_job_units', 2)

    - update multiple settings at once::

        >>> project.settings.update({'default_job_units': 1,
        ...                          'job_runtime_limit': 20})

    - delete project setting by name::

        >>> project.settings.delete('job_runtime_limit')
    """
    def set(self, key, value):
        """Update project setting value by key.

        :param key: a string setting key.
        :param value: new setting value.
        """
        # FIXME drop the method when post-by-key is implemented on server side
        self.update({key: value})



from __future__ import absolute_import

import json
import logging

from .proxy import _ItemsResourceProxy, _DownloadableProxyMixin
from .utils import LogLevel


class Logs(_ItemsResourceProxy, _DownloadableProxyMixin):
    """Representation of collection of job logs.

    Not a public constructor: use :class:`~scrapinghub.client.jobs.Job` instance
    to get a :class:`Logs` instance. See :attr:`~scrapinghub.client.jobs.Job.logs`
    attribute.

    Please note that :meth:`list` method can use a lot of memory and for a
    large amount of logs it's recommended to iterate through it via :meth:`iter`
    method (all params and available filters are same for both methods).

    Usage:

    - retrieve all logs from a job::

        >>> job.logs.iter()
        <generator object mpdecode at 0x10f5f3aa0>

    - iterate through first 100 log entries and print them::

        >>> for log in job.logs.iter(count=100):
        ...     print(log)

    - retrieve a single log entry from a job::

        >>> job.logs.list(count=1)
        [{
            'level': 20,
            'message': '[scrapy.core.engine] Closing spider (finished)',
            'time': 1482233733976,
        }]

    - retrive logs with a given log level and filter by a word::

        >>> filters = [("message", "contains", ["mymessage"])]
        >>> job.logs.list(level='WARNING', filter=filters)
        [{
            'level': 30,
            'message': 'Some warning: mymessage',
            'time': 1486375511188,
        }]
    """
    def log(self, message, level=logging.INFO, ts=None, **other):
        """Base method to write a log entry.

        :param message: a string message.
        :param level: (optional) logging level, default to INFO.
        :param ts: (optional) UNIX timestamp in milliseconds.
        :param \*\*other: other optional kwargs.
        """
        self._origin.log(message, level=level, ts=ts, **other)

    def debug(self, message, **other):
        """Log a message with DEBUG level."""
        self._origin.debug(message, **other)

    def info(self, message, **other):
        """Log a message with INFO level."""
        self._origin.info(message, **other)

    def warn(self, message, **other):
        """Log a message with WARN level."""
        self._origin.warn(message, **other)
    warning = warn

    def error(self, message, **other):
        """Log a message with ERROR level."""
        self._origin.error(message, **other)

    def batch_write_start(self):
        """Override to set a start parameter when commencing writing."""
        return self._origin.batch_write_start()

    def _modify_iter_params(self, params):
        """Modify iter() filters on-the-fly.

        - convert offset to start parameter
        - check log level and create a corresponding meta filter

        :param params: an original dictionary with params.
        :return: a modified dictionary with params.
        :rtype: :class:`dict`
        """
        params = super(Logs, self)._modify_iter_params(params)
        offset = params.pop('offset', None)
        if offset:
            params['start'] = '{}/{}'.format(self.key, offset)
        level = params.pop('level', None)
        if level:
            minlevel = getattr(LogLevel, level, None)
            if minlevel is None:
                raise ValueError("Unknown log level: {}".format(level))
            level_filter = json.dumps(['level', '>=', [minlevel]])
            # there can already be some filters handled by super class method
            params['filter'] = params.get('filter', []) + [level_filter]
        return params



import time
import socket
import random
import logging
import warnings
import six
from six.moves import range
from six.moves.queue import Queue
from io import BytesIO
from gzip import GzipFile
from itertools import count
import requests
from collections import deque
from threading import Thread, Event
from .utils import xauth, iterqueue, sizeof_fmt
from .serialization import jsonencode

logger = logging.getLogger('hubstorage.batchuploader')


class BatchUploader(object):

    # Wait time between all batches status checks
    worker_loop_delay = 1.0
    # Max number of retry attempts before giving up.
    worker_max_retries = 200
    # The delay increases exponentially with the number of attempts but is
    # bounded with these values on both sides.
    worker_min_interval = 30
    worker_max_interval = 600
    # Each delay is also randomized by multiplying by random(0.5, 1.5), so the
    # total delay using the current parameters is an almost gaussian random
    # number with the following characteristics (see Irwin-Hall distribution):
    #
    # - average = 30hrs
    # - minimum = 15hrs
    # - maximum = 45hrs
    # - standard deviation = approx. 40m, which means that 95% of the time the
    #   total delay will be within 2*std = 1h20m of the average.

    def __init__(self, client):
        self.client = client
        self.closed = False
        self._wait_event = Event()
        self._writers = deque()
        self._thread = Thread(target=self._worker)
        self._thread.daemon = True
        self._thread.start()

    def create_writer(self, url, start=0, auth=None, size=1000, interval=15,
                      qsize=None, content_encoding='identity',
                      maxitemsize=1024 ** 2, callback=None):
        # callback shouldn't try to inject more items in the queue
        # otherwise it can lead to deadlock on _checkpoint step
        assert not self.closed, 'Can not create new writers when closed'
        auth = xauth(auth) or self.client.auth
        w = _BatchWriter(url=url,
                         auth=auth,
                         size=size,
                         start=start,
                         interval=interval,
                         qsize=qsize,
                         maxitemsize=maxitemsize,
                         content_encoding=content_encoding,
                         uploader=self,
                         callback=callback)
        self._writers.append(w)
        return w

    def close(self, timeout=None):
        self.closed = True
        self.interrupt()
        self._thread.join(timeout)

    def interrupt(self):
        self._wait_event.set()

    def __del__(self):
        if not self.closed:
            warnings.warn("%r not closed properly, some items may have been "
                          "lost!: %r" % (self.__class__.__name__, self._writers))

    def _interruptable_sleep(self):
        self._wait_event.wait(self.worker_loop_delay)
        self._wait_event.clear()

    def _worker(self):
        ctr = count()
        while True:
            if not self._writers:
                # Stop thread if closed and idle, but if open wait for writers
                if self.closed:
                    break
                self._interruptable_sleep()
                continue

            # Delay once all writers are processed
            if (next(ctr) % len(self._writers) == 0) and not self.closed:
                self._interruptable_sleep()

            # Get next writer to process
            w = self._writers.popleft()

            # Close open writers if uploader is closed
            if self.closed and not w.closed:
                w.close(block=False)

            # Checkpoint writer if eligible
            now = time.time()
            if w.itemsq.qsize() >= w.size or w.closed or w.flushme \
                    or w.checkpoint < now - w.interval:
                self._checkpoint(w)
                w.checkpoint = now

            # Re-queue pending or open writers
            if not (w.closed and w.itemsq.empty()):
                self._writers.append(w)

    def _checkpoint(self, w):
        q = w.itemsq
        qiter = iterqueue(q, w.size)
        data = self._content_encode(qiter, w)
        if qiter.count > 0:
            response = self._tryupload({
                'url': w.url,
                'offset': w.offset,
                'data': data,
                'auth': w.auth,
                'content-encoding': w.content_encoding,
            })
            w.offset += qiter.count
            if w.callback is not None:
                try:
                    w.callback(response)
                except Exception:
                    logger.exception("Callback for %s failed", w.url)
            for _ in range(qiter.count):
                q.task_done()

    def _content_encode(self, qiter, w):
        ce = w.content_encoding
        if ce == 'identity':
            return _encode_identity(qiter)
        elif ce == 'gzip':
            return _encode_gzip(qiter)
        else:
            raise ValueError('Writer using unknown content encoding: %s' % ce)

    def _tryupload(self, batch):
        """Retry uploads in case of server failures

        Use polinomial backoff with 10 minutes maximum interval that accounts
        for ~30 hours of total retry time.

        >>> sum(min(x**2, 600) for x in range(200)) / 3600
        30
        """
        url = batch['url']
        offset = batch['offset']
        for retryn in range(self.worker_max_retries):
            emsg = ''
            try:
                r = self._upload(batch)
                r.raise_for_status()
                if not (200 <= r.status_code < 300):
                    logger.warning('Discarding write to url=%s offset=%s: '
                                   '[HTTP error %s] %s\n%s', url, offset,
                                   r.status_code, r.reason, r.text.rstrip())
                return r
            except (socket.error, requests.RequestException) as e:
                if isinstance(e, requests.HTTPError):
                    emsg = "[HTTP error {0}] {1}".format(e.response.status_code,
                                                         e.response.text.rstrip())
                else:
                    emsg = str(e)
                logger.info("Retrying url=%s offset=%s: %s", url, offset, emsg)
            except Exception:
                logger.exception('Non retryable failure on url=%s offset=%s',
                                 url, offset)
                break

            backoff = min(max(retryn ** 2, self.worker_min_interval),
                          self.worker_max_interval)
            time.sleep(backoff * (0.5 + random.random()))

    def _upload(self, batch):
        params = {'start': batch['offset']}
        headers = {'content-encoding': batch['content-encoding']}
        return self.client.session.request(
            method='POST',
            url=batch['url'],
            data=batch['data'],
            auth=batch['auth'],
            timeout=self.client.connection_timeout,
            params=params,
            headers=headers,
        )

class ValueTooLarge(ValueError):
    """Raised when a serialized item is greater than 1MB"""


class _BatchWriter(object):
    #: Truncate overly big items to that many bytes for the error message.
    ERRMSG_DATA_TRUNCATION_LEN = 1024

    def __init__(self, url, start, auth, size, interval, qsize,
                 maxitemsize, content_encoding, uploader, callback=None):
        self.url = url
        self.offset = start
        self._nextid = count(start)
        self.auth = auth
        self.size = size
        self.interval = interval
        self.maxitemsize = maxitemsize
        self.content_encoding = content_encoding
        self.checkpoint = time.time()
        self.itemsq = Queue(size * 2 if qsize is None else qsize)
        self.closed = False
        self.flushme = False
        self.uploader = uploader
        self.callback = callback

    def write(self, item):
        assert not self.closed, 'attempting writes to a closed writer'
        data = jsonencode(item)
        if len(data) > self.maxitemsize:
            truncated_data = data[:self.ERRMSG_DATA_TRUNCATION_LEN] + "..."
            raise ValueTooLarge(
                'Value exceeds max encoded size of {}: {!r}'
                .format(sizeof_fmt(self.maxitemsize), truncated_data))

        self.itemsq.put(data)
        if self.itemsq.full():
            self.uploader.interrupt()
        return next(self._nextid)

    def flush(self):
        self.flushme = True
        self._waitforq()
        self.flushme = False

    def close(self, block=True):
        self.closed = True
        if block:
            self._waitforq()

    def _waitforq(self):
        self.uploader.interrupt()
        self.itemsq.join()

    def __str__(self):
        return self.url


def _encode_identity(iterable):
    data = BytesIO()
    for item in iterable:
        if isinstance(item, six.text_type):
            item = item.encode('utf8')
        data.write(item)
        data.write(b'\n')
    return data.getvalue()


def _encode_gzip(iterable):
    data = BytesIO()
    with GzipFile(fileobj=data, mode='w') as gzo:
        for item in iterable:
            if isinstance(item, six.text_type):
                item = item.encode('utf8')
            gzo.write(item)
            gzo.write(b'\n')
    return data.getvalue()



from __future__ import absolute_import

from .proxy import _ItemsResourceProxy, _DownloadableProxyMixin


class Items(_ItemsResourceProxy, _DownloadableProxyMixin):
    """Representation of collection of job items.

    Not a public constructor: use :class:`~scrapinghub.client.jobs.Job`
    instance to get a :class:`Items` instance. See
    :attr:`~scrapinghub.client.jobs.Job.items` attribute.

    Please note that :meth:`list` method can use a lot of memory and for
    a large amount of logs it's recommended to iterate through it via
    :meth:`iter` method (all params and available filters are same for
    both methods).

    Usage:

    - retrieve all scraped items from a job::

        >>> job.items.iter()
        <generator object mpdecode at 0x10f5f3aa0>

    - iterate through first 100 items and print them::

        >>> for log in job.logs.iter(count=100):
        ...     print(log)

    - retrieve items with timestamp greater or equal to given timestamp
      (item here is an arbitrary dictionary depending on your code)::

        >>> job.items.list(startts=1447221694537)
        [{
            'name': ['Some custom item'],
            'url': 'http://some-url/item.html',
            'size': 100000,
        }]

    - retrieve 1 item with multiple filters::

        >>> filters = [("size", ">", [30000]), ("size", "<", [40000])]
        >>> job.items.list(count=1, filter=filters)
        [{
            'name': ['Some other item'],
            'url': 'http://some-url/other-item.html',
            'size': 35000,
        }]
    """

    def _modify_iter_params(self, params):
        """Modify iter filter to convert offset to start parameter.

        :return: a dict with updated set of params.
        :rtype: :class:`dict`
        """
        params = super(Items, self)._modify_iter_params(params)
        offset = params.pop('offset', None)
        if offset:
            params['start'] = '{}/{}'.format(self.key, offset)
        return params



from scrapinghub import Connection as _Connection
from scrapinghub import HubstorageClient as _HubstorageClient

from .exceptions import _wrap_http_errors
from .projects import Projects
from .utils import parse_auth
from .utils import parse_project_id, parse_job_key


__all__ = ['ScrapinghubClient']


class Connection(_Connection):

    @_wrap_http_errors
    def _request(self, *args, **kwargs):
        return super(Connection, self)._request(*args, **kwargs)


class HubstorageClient(_HubstorageClient):

    @_wrap_http_errors
    def request(self, *args, **kwargs):
        return super(HubstorageClient, self).request(*args, **kwargs)


class ScrapinghubClient(object):
    """Main class to work with Scrapinghub API.

    :param auth: Scrapinghub APIKEY or other SH auth credentials.
    :param dash_endpoint: (optional) Scrapinghub Dash panel url.
    :param \*\*kwargs: (optional) Additional arguments for
        :class:`~scrapinghub.hubstorage.HubstorageClient` constructor.

    :ivar projects: projects collection,
        :class:`~scrapinghub.client.projects.Projects` instance.

    Usage::

        >>> from scrapinghub import ScrapinghubClient
        >>> client = ScrapinghubClient('APIKEY')
        >>> client
        <scrapinghub.client.ScrapinghubClient at 0x1047af2e8>
    """

    def __init__(self, auth=None, dash_endpoint=None, **kwargs):
        self.projects = Projects(self)
        login, password = parse_auth(auth)
        self._connection = Connection(apikey=login,
                                      password=password,
                                      url=dash_endpoint)
        self._hsclient = HubstorageClient(auth=(login, password), **kwargs)

    def get_project(self, project_id):
        """Get :class:`scrapinghub.client.projects.Project` instance with
        a given project id.

        The method is a shortcut for client.projects.get().

        :param project_id: integer or string numeric project id.
        :return: a project instance.
        :rtype: :class:`~scrapinghub.client.projects.Project`

        Usage::

            >>> project = client.get_project(123)
            >>> project
            <scrapinghub.client.projects.Project at 0x106cdd6a0>
        """
        return self.projects.get(parse_project_id(project_id))

    def get_job(self, job_key):
        """Get :class:`~scrapinghub.client.jobs.Job` with a given job key.

        :param job_key: job key string in format ``project_id/spider_id/job_id``,
            where all the components are integers.
        :return: a job instance.
        :rtype: :class:`~scrapinghub.client.jobs.Job`

        Usage::

            >>> job = client.get_job('123/1/1')
            >>> job
            <scrapinghub.client.jobs.Job at 0x10afe2eb1>
        """
        project_id = parse_job_key(job_key).project_id
        return self.projects.get(project_id).jobs.get(job_key)

    def close(self, timeout=None):
        """Close client instance.

        :param timeout: (optional) float timeout secs to stop gracefully.
        """
        self._hsclient.close(timeout=timeout)



from __future__ import absolute_import
from functools import partial
from collections import defaultdict

from six import string_types

from ..hubstorage.frontier import Frontier as _Frontier
from ..hubstorage.utils import urlpathjoin

from .proxy import _Proxy
from .utils import update_kwargs


class _HSFrontier(_Frontier):
    """Modified hubstorage Frontier with newcount per slot."""

    def __init__(self, *args, **kwargs):
        super(_HSFrontier, self).__init__(*args, **kwargs)
        self.newcount = defaultdict(int)

    def _get_writer(self, frontier, slot):
        """Modified helper method to create a batchuploader writer with updated
        callback to write newcount data per slot.

        :return: a batchuploader writer instance.
        :rtype: :class:`~scrapinghub.hubstorage.batchuploader._BatchWriter`
        """
        key = (frontier, slot)
        writer = self._writers.get(key)
        if not writer:
            writer = self.client.batchuploader.create_writer(
                url=urlpathjoin(self.url, frontier, 's', slot),
                auth=self.auth,
                size=self.batch_size,
                start=self.batch_start,
                interval=self.batch_interval,
                qsize=self.batch_qsize,
                content_encoding=self.batch_content_encoding,
                callback=partial(self._writer_callback, key),
            )
            self._writers[key] = writer
        return writer

    def _writer_callback(self, key, response):
        """Writer callback function when new batch is added."""
        self.newcount[key] += response.json()["newcount"]


class Frontiers(_Proxy):
    """Frontiers collection for a project.

    Not a public constructor: use :class:`~scrapinghub.client.projects.Project`
    instance to get a :class:`Frontiers` instance.
    See :attr:`~scrapinghub.client.Project.frontiers` attribute.

    Usage:

    - get all frontiers from a project::

        >>> project.frontiers.iter()
        <list_iterator at 0x103c93630>

    - list all frontiers::

        >>> project.frontiers.list()
        ['test', 'test1', 'test2']

    - get a frontier by name::

        >>> project.frontiers.get('test')
        <scrapinghub.client.frontiers.Frontier at 0x1048ae4a8>

    - flush data of all frontiers of a project::

        >>> project.frontiers.flush()

    - show amount of new requests added for all frontiers::

        >>> project.frontiers.newcount
        3

    - close batch writers of all frontiers of a project::

        >>> project.frontiers.close()
    """
    def __init__(self, *args, **kwargs):
        super(Frontiers, self).__init__(*args, **kwargs)

    def get(self, name):
        """Get a frontier by name.

        :param name: a frontier name string.
        :return: a frontier instance.
        :rtype: :class:`Frontier`
        """
        return Frontier(self._client, self, name)

    def iter(self):
        """Iterate through frontiers.

        :return: an iterator over frontiers names.
        :rtype: :class:`collections.Iterable[str]`
        """
        return iter(self.list())

    def list(self):
        """List frontiers names.

        :return: a list of frontiers names.
        :rtype: :class:`list[str]`
        """
        return next(self._origin.apiget('list'))

    @property
    def newcount(self):
        """Integer amount of new entries added to all frontiers."""
        return sum(self._origin.newcount.values())

    def flush(self):
        """Flush data in all frontiers writer threads."""
        self._origin.flush()

    def close(self):
        """Close frontier writer threads one-by-one."""
        self._origin.close()


class Frontier(object):
    """Representation of a frontier object.

    Not a public constructor: use :class:`Frontiers` instance to get a
    :class:`Frontier` instance. See :meth:`Frontiers.get` method.

    Usage:

    - get iterator with all slots::

        >>> frontier.iter()
        <list_iterator at 0x1030736d8>

    - list all slots::

        >>> frontier.list()
        ['example.com', 'example.com2']

    - get a slot by name::

        >>> frontier.get('example.com')
        <scrapinghub.client.frontiers.FrontierSlot at 0x1049d8978>

    - flush frontier data::

        >>> frontier.flush()

    - show amount of new requests added to frontier::

        >>> frontier.newcount
        3
    """
    def __init__(self, client, frontiers, name):
        self.key = name
        self._client = client
        self._frontiers = frontiers

    def get(self, slot):
        """Get a slot by name.

        :return: a frontier slot instance.
        :rtype: :class:`FrontierSlot`
        """
        return FrontierSlot(self._client, self, slot)

    def iter(self):
        """Iterate through slots.

        :return: an iterator over frontier slots names.
        :rtype: :class:`collections.Iterable[str]`
        """
        return iter(self.list())

    def list(self):
        """List all slots.

        :return: a list of frontier slots names.
        :rtype: :class:`list[str]`
        """
        return next(self._frontiers._origin.apiget((self.key, 'list')))

    def flush(self):
        """Flush data for a whole frontier."""
        writers = self._frontiers._origin._writers
        for (fname, _), writer in writers.items():
            if fname == self.key:
                writer.flush()

    @property
    def newcount(self):
        """Integer amount of new entries added to frontier."""
        newcount_values = self._frontiers._origin.newcount
        return sum(v for (frontier, _), v in newcount_values.items()
                   if frontier == self.key)


class FrontierSlot(object):
    """Representation of a frontier slot object.

    Not a public constructor: use :class:`Frontier` instance to get a
    :class:`FrontierSlot` instance. See :meth:`Frontier.get` method.

    Usage:

    - add request to a queue::

        >>> data = [{'fp': 'page1.html', 'p': 1, 'qdata': {'depth': 1}}]
        >>> slot.q.add('example.com', data)

    - add fingerprints to a slot::

        >>> slot.f.add(['fp1', 'fp2'])

    - flush data for a slot::

        >>> slot.flush()

    - show amount of new requests added to a slot::

        >>> slot.newcount
        2

    - read requests from a slot::

        >>> slot.q.iter()
        <generator object jldecode at 0x1049aa9e8>
        >>> slot.q.list()
        [{'id': '0115a8579633600006',
          'requests': [['page1.html', {'depth': 1}]]}]

    - read fingerprints from a slot::

        >>> slot.f.iter()
        <generator object jldecode at 0x103de4938>
        >>> slot.f.list()
        ['page1.html']

    - delete a batch with requests from a slot::

        >>> slot.q.delete('0115a8579633600006')

    - delete a whole slot::

        >>> slot.delete()
    """
    def __init__(self, client, frontier, slot):
        self.key = slot
        self._client = client
        self._frontier = frontier
        self.fingerprints = FrontierSlotFingerprints(self)
        self.queue = FrontierSlotQueue(self)

    @property
    def f(self):
        """Shortcut to have quick access to slot fingerprints.

        :return: fingerprints collection for the slot.
        :rtype: :class:`FrontierSlotFingerprints`
        """
        return self.fingerprints

    @property
    def q(self):
        """Shortcut to have quick access to a slot queue.

        :return: queue instance for the slot.
        :rtype: :class:`FrontierSlotQueue`
        """
        return self.queue

    def delete(self):
        """Delete the slot."""
        origin = self._frontier._frontiers._origin
        origin.delete_slot(self._frontier.key, self.key)
        origin.newcount.pop((self._frontier.key, self.key), None)

    def flush(self):
        """Flush data for the slot."""
        writers = self._frontier._frontiers._origin._writers
        writer = writers.get((self._frontier.key, self.key))
        if writer:
            writer.flush()

    @property
    def newcount(self):
        """Integer amount of new entries added to slot."""
        newcount_values = self._frontier._frontiers._origin.newcount
        return newcount_values.get((self._frontier.key, self.key), 0)


class FrontierSlotFingerprints(object):
    """Representation of request fingerprints collection stored in slot."""

    def __init__(self, slot):
        self.key = slot.key
        self._frontier = slot._frontier
        self._slot = slot

    def add(self, fps):
        """Add new fingerprints to slot.

        :param fps: a list of string fingerprints to add.
        """
        origin = self._frontier._frontiers._origin
        writer = origin._get_writer(self._frontier.key, self.key)
        fps = list(fps) if not isinstance(fps, list) else fps
        if not all(isinstance(fp, string_types) for fp in fps):
            raise ValueError('Fingerprint should be of a string type')
        for fp in fps:
            writer.write({'fp': fp})

    def iter(self, **params):
        """Iterate through fingerprints in the slot.

        :param \*\*params: (optional) additional query params for the request.
        :return: an iterator over fingerprints.
        :rtype: :class:`collections.Iterable[str]`
        """
        origin = self._frontier._frontiers._origin
        path = (self._frontier.key, 's', self.key, 'f')
        for fp in origin.apiget(path, params=params):
            yield fp.get('fp')

    def list(self, **params):
        """List fingerprints in the slot.

        :param \*\*params: (optional) additional query params for the request.
        :return: a list of fingerprints.
        :rtype: :class:`list[str]`
        """
        return list(self.iter(**params))


class FrontierSlotQueue(object):
    """Representation of request batches queue stored in slot."""

    def __init__(self, slot):
        self.key = slot.key
        self._frontier = slot._frontier
        self._slot = slot

    def add(self, fps):
        """Add requests to the queue."""
        origin = self._frontier._frontiers._origin
        return origin.add(self._frontier.key, self.key, fps)

    def iter(self, mincount=None, **params):
        """Iterate through batches in the queue.

        :param mincount: (optional) limit results with min amount of requests.
        :param \*\*params: (optional) additional query params for the request.
        :return: an iterator over request batches in the queue where each
            batch is represented with a dict with ('id', 'requests') field.
        :rtype: :class:`collections.Iterable[dict]`
        """
        origin = self._frontier._frontiers._origin
        path = (self._frontier.key, 's', self.key, 'q')
        update_kwargs(params, mincount=mincount)
        return origin.apiget(path, params=params)

    def list(self, mincount=None, **params):
        """List request batches in the queue.

        :param mincount: (optional) limit results with min amount of requests.
        :param \*\*params: (optional) additional query params for the request.
        :return: a list of request batches in the queue where each batch
            is represented with a dict with ('id', 'requests') field.
        :rtype: :class:`list[dict]`
        """
        return list(self.iter(mincount=mincount, **params))

    def delete(self, ids):
        """Delete request batches from the queue."""
        origin = self._frontier._frontiers._origin
        return origin.delete(self._frontier.key, self.key, ids)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from functools import wraps

from requests import HTTPError

from ..legacy import APIError


def _get_http_error_msg(exc):
    if isinstance(exc, HTTPError):
        try:
            payload = exc.response.json()
        except ValueError:
            payload = None
        if payload and isinstance(payload, dict):
            message = payload.get('message')
            if message:
                return message
        elif exc.response.text:
            return exc.response.text
    return str(exc)


class ScrapinghubAPIError(Exception):
    """Base exception class."""

    def __init__(self, message=None, http_error=None):
        self.http_error = http_error
        if not message:
            message = _get_http_error_msg(http_error)
        super(ScrapinghubAPIError, self).__init__(message)


class BadRequest(ScrapinghubAPIError):
    """Usually raised in case of 400 response from API."""


class Unauthorized(ScrapinghubAPIError):
    """Request lacks valid authentication credentials for the target resource."""


class NotFound(ScrapinghubAPIError):
    """Entity doesn't exist (e.g. spider or project)."""


class ValueTooLarge(ScrapinghubAPIError):
    """Value cannot be writtent because it exceeds size limits."""


class DuplicateJobError(ScrapinghubAPIError):
    """Job for given spider with given arguments is already scheduled or running."""


class ServerError(ScrapinghubAPIError):
    """Indicates some server error: something unexpected has happened."""


def _wrap_http_errors(method):
    """Internal helper to handle exceptions gracefully."""
    @wraps(method)
    def wrapped(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except HTTPError as exc:
            status_code = exc.response.status_code
            if status_code == 400:
                raise BadRequest(http_error=exc)
            elif status_code == 401:
                raise Unauthorized(http_error=exc)
            elif status_code == 404:
                raise NotFound(http_error=exc)
            elif status_code == 413:
                raise ValueTooLarge(http_error=exc)
            elif 400 <= status_code < 500:
                raise ScrapinghubAPIError(http_error=exc)
            elif 500 <= status_code < 600:
                raise ServerError(http_error=exc)
            raise
        except APIError as exc:
            msg = exc.args[0]
            if exc._type == APIError.ERR_NOT_FOUND:
                raise NotFound(msg)
            elif exc._type == APIError.ERR_VALUE_ERROR:
                raise ValueError(msg)
            elif exc._type == APIError.ERR_BAD_REQUEST:
                raise BadRequest(msg)
            elif exc._type == APIError.ERR_AUTH_ERROR:
                raise Unauthorized(http_error=exc)
            elif exc._type == APIError.ERR_SERVER_ERROR:
                raise ServerError(http_error=exc)
            raise ScrapinghubAPIError(msg)
    return wrapped
